from marshmallow import Schema, fields
from stschema.util.base_modules import (CookieSchema,
                                        ManufacturerSchema,
                                        DeviceContextSchema)


class DeviceDiscoverySchema(Schema):
    """
    The DeviceDiscoverySchema handles
    the serialization of the SchemaDevice
    class supporting the nested
    CookieSchema, ManufacturerSchema, and
    DeviceContextSchema.
    It converts Snake Case attributes
    to Camel Case format following REST
    formatting conventions for JSON
    string objects.
    """

    externalDeviceId = fields.Field(attribute='external_device_id', required=True)
    friendlyName = fields.Field(attribute='friendly_name', required=True)
    deviceHandlerType = fields.Field(attribute='device_handler_type', required=True)
    deviceUniqueId = fields.Field(attribute='device_unique_id')
    deviceCookie = fields.Nested(CookieSchema, attribute='device_cookie')
    manufacturerInfo = fields.Nested(ManufacturerSchema, attribute='manufacturer_info', required=True)
    deviceContext = fields.Nested(DeviceContextSchema, attribute='device_context')



from marshmallow import Schema, fields
from stschema.util.base_modules import (
    CallbackAuthentication,
    CallbackAuthSchema,
    HeadersSchema,
    BaseResponse
)


class AccessTokenRequest(BaseResponse):
    """
    The AccessTokenRequest class handles the
    information necessary to request or refresh
    Access Tokens at the SmartThings Cloud.
    """
    def __init__(self,
                 client_id: str,
                 client_secret: str,
                 code: str,
                 refresh_token: str,
                 request_id: str,
                 grant_type: str='authorization_code') -> 'AccessTokenRequest':
        BaseResponse.__init__(self, interaction_type='accessTokenRequest', request_id=request_id)

        # Refresh Token reference
        if not code and refresh_token:
            grant_type = 'refresh_token'
            self.headers.interaction_type = 'refreshAccessTokens'

        self.callback_authentication = CallbackAuthentication(
            client_id = client_id,
            client_secret = client_secret,
            code = code,
            refresh_token = refresh_token,
            grant_type = grant_type)


class AccessTokenRequestSchema(Schema):
    """
    The AccessTokenRequestSchema handles the
    serialization of AccessTokenRequest
    class.
    It converts Snake Case attributes
    to Camel Case format following REST
    formatting conventions for JSON
    string objects.
    """
    headers = fields.Nested(HeadersSchema, attribute='headers', required=True)
    callbackAuthentication = fields.Nested(CallbackAuthSchema, attribute='callback_authentication', required=True)



import logging


class EventLogger:
    def __init__(self, enable: bool):
        self.logger = self._logger(enable)

    def _logger(self, enable):
        logger = logging.getLogger('[stschema]')
        # Handle levels
        logger.setLevel(logging.WARNING)
        if enable:
            logger.setLevel(logging.DEBUG)

        #  Stream handler instance.
        stream  = logging.StreamHandler()
        # Formatter instance
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(message)s')
        stream.setFormatter(formatter)
        logger.addHandler(stream)
        logger.removeHandler(stream)
        return logger



from marshmallow import Schema, fields


class BaseHeaders:
    """
    The BaseHeaders represents the
    communication metadata between a
    third-party Cloud and the SmartThings
    Cloud. 'schema' and 'version' attributes
    are default class values:
        schema = 'st-schema'
        version = '1.0'

        :::param interaction_type
        :::param request_id
    """

    schema = 'st-schema'
    version = '1.0'

    def __init__(self, interaction_type: str, request_id: str) -> 'BaseHeaders':
        self.interaction_type = interaction_type
        self.request_id = request_id


class HeadersSchema(Schema):
    """
    The HeaderSchema handles the
    serialization of the BaseHeaders
    class.
    It converts Snake Case attributes
    to Camel Case format following REST
    formatting conventions for JSON
    string objects.
    """

    schema = fields.Str()
    version = fields.Str()
    interactionType = fields.Str(attribute='interaction_type')
    requestId = fields.Str(attribute='request_id')



from .discovery_response import DiscoveryResponse, DiscoveryResponseSchema
from .state_refresh_response import StateResponse, StateRefreshResponseSchema
from .global_error_response import GlobalErrorResponse, GlobalErrorSchema



"""
The SchemaConnector class provides a json_data
handler that will take the entity-body of a
Webhook Http Request and process it according to
the Interaction Type passed. It is intended to be
used with JSON data only.

Also, it provides a series of Interaction Type
handler interfaces that will parse the JSON body
and pass the necessary information to simplify
the development of Schema Connector instances.
"""
from stschema.schema_response import SchemaResponse
from stschema.util import EventLogger, ErrorHandler


class SchemaConnector(SchemaResponse):
    """
    The Schema Connector class provides
    a developer interface to control the flow
    of Interaction Types through interaction
    resource handlers.
        :::param enable_logger: boolean
    """
    def __init__(self, enable_logger: bool=False):
        self.logger = EventLogger(enable_logger).logger
        self.error_handler = ErrorHandler()

    def interaction_handler(self, json_data: dict):
        """
        This method helps with the
        flow-control of Interaction Types
        and returns interaction resource
        handlers respectively.
            :::param json_data: dict
        """

        if not isinstance(json_data, dict):
            raise TypeError('dict argument expected, not (%s, %s)' % (type(json_data), json_data))
        else:
            # Validating SmartThings Schema Headers
            # information that contains critical
            # values of requests persinstence.
            #   - Request Id
            #   - Interaction Type reference.
            headers_args = json_data.get('headers')
            # Validating authorization attribute
            # that contains the access token issued
            # by the OAuth2.0 server.
            auth_args = json_data.get('authentication')
            if not headers_args:
                raise AttributeError('missing or null "headers" attribute at json_data')
            elif not auth_args:
                raise AttributeError('missing or null "authentication" attribute at json_data')
            else:
                request_id_arg = headers_args.get('requestId')
                interaction_type_arg = headers_args.get('interactionType')
                if not request_id_arg:
                    raise AttributeError('missing or null "requestId" attribute at json_data.headers')
                elif not interaction_type_arg:
                    raise AttributeError('missing or null "interactionType" attribute at json_data.headers')
                else:
                    # Logging respective json_data
                    self.logger.info('[%s] - %s' % (interaction_type_arg, json_data))
                    return self._interaction_handler(json_data)

    def _interaction_handler(self, data):
        data_headers = data['headers']
        interaction_type = data_headers['interactionType']
        request_id = data_headers['requestId']
        access_token = data['authentication']['token']
        # Handle interaction type
        if interaction_type == 'discoveryRequest':
            return self.discovery_handler(request_id, access_token)
        elif interaction_type == 'stateRefreshRequest':
            return self.state_refresh_handler(data['devices'], request_id, access_token)
        elif interaction_type == 'commandRequest':
            return self.command_handler(data['devices'], request_id, access_token)
        elif interaction_type == 'grantCallbackAccess':
            return self.grant_callback_access(data['callbackAuthentication'], data['callbackUrls'])
        elif interaction_type == 'integrationDeleted':
            return self.integration_deleted(data['callbackAuthentication'])
        elif interaction_type == 'interactionResult':
            # Since detail of Interaction Types
            # are dynamic, instead of complex
            # flow-control conditions, headers
            # and authentication will be poped
            # before calling handler.
            if data.get('authentication'):
                data.pop('authentication')
            data.pop('headers')
            # Originating Interaction Type
            origin = data.get('originatingInteractionType')
            return self.interaction_result_handler(data, origin)

    def discovery_handler(self, request_id: str, access_token: str):
        """
        Implementation example handling
        specific attributes from json_data:

            discovery_handler(self, request_id, access_token):
                # Custom implementation here...
                return super().discovery_response(devices, request_id)
            ...
        """
        return self.error_handler.not_implemented_error('discovery_handler')

    def state_refresh_handler(self, devices: list, request_id: str, access_token: str):
        """
        Implementation example handling
        specific attributes from json_data:

            state_refresh_handler(self, devices, request_id, access_token):
                # Custom implementation here...
                return super().state_refresh_response(devices, request_id)
        """
        return self.error_handler.not_implemented_error('state_refresh_handler')

    def command_handler(self, devices: list, request_id: str, access_token: str):
        """
        Implementation example handling
        specific attributes from json_data:

            command_handler(self, devices, request_id, access_token):
                # Custom implementation here...
                return super().command_request(devices, request_id)
        """
        return self.error_handler.not_implemented_error('command_handler')

    def grant_callback_access(self, callback_authentication: dict, callback_urls: dict):
        """
        Implementation example handling
        specific attributes from json_data:

            grant_callback_access(self, callback_authentication):
            ...
        """
        return self.error_handler.not_implemented_error('grant_callback_access')

    def integration_deleted(self, callback_authentication: dict):
        """
        Implementation example handling
        specific attributes from json_data:

            integration_deleted(self, callback_authentication):
            ...
        """
        return self.error_handler.not_implemented_error('integration_deleted')

    def interaction_result_handler(self, interaction_result: dict, origin: str):
        """
        Implementation example handling
        specific attributes from json_data:

            interaction_result_handler(self, devices, request_id):
            ...
        """
        return self.error_handler.not_implemented_error('interaction_result_handler')



from marshmallow import Schema, fields


class BaseState:
    """
    The BaseState uses the capability's
    metadata to represent a device's state.
        :::param component
        :::param capability
        :::param attribute
        :::param value
        :::param unit
    """

    def __init__(self, capability: str, attribute: str, value, unit: str, component: str) -> 'BaseState':
        self.component = component
        self.capability = capability
        self.attribute = attribute
        self.value = value
        if unit:
            self.unit = unit


class StateSchema(Schema):
    """
    The StateSchema handles the
    serialization of the BaseState
    class.
    It converts Snake Case attributes
    to Camel Case format following REST
    formatting conventions for JSON
    string objects.
    """

    component = fields.Str()
    capability = fields.Str()
    attribute = fields.Str()
    value = fields.Field()
    unit = fields.Str()



import enum


class StateErrorEnum(enum.Enum):
    """
    The StateErrorEnum represents
    the supported device error states
    enumerators.
    """

    CAPABILITY_NOT_SUPPORTED = 'CAPABILITY-NOT-SUPPORTED'
    DEVICE_DELETED = 'DEVICE-DELETED'
    DEVICE_UNAVAILABLE = 'DEVICE-UNAVAILABLE'
    RESOURCE_CONSTRAINT_VIOLATION = 'RESOURCE-CONSTRAINT-VIOLATION'


class GlobalErrorEnum(enum.Enum):
    """
    The GlobalErrorEnum represents
    the supported global-context error
    enumerators.
    """

    BAD_REQUEST = 'BAD-REQUEST'
    INTEGRATION_DELETED = 'INTEGRATION-DELETED'
    INVALID_CLIENT = 'INVALID-CLIENT'
    INVALID_CLIENT_SECRET = 'INVALID-CLIENT-SECRET'
    INVALID_CODE = 'INVALID-CODE'
    INVALID_INTERACTION_TYPE = 'INVALID-INTERACTION-TYPE'
    INVALID_TOKEN = 'INVALID-TOKEN'
    TOKEN_EXPIRED = 'TOKEN-EXPIRED'
    UNSUPPORTED_GRANT_TYPE = 'UNSUPPORTED-GRANT-TYPE'



import requests
from stschema import SchemaDevice
# Schemas
from stschema.schema_response.responses import (
    StateRefreshResponseSchema,
    DiscoveryResponseSchema
)
# Callback interaction type objects and schemas
from stschema.schema_callbacks.callbacks import (
    AccessTokenRequest,
    AccessTokenRequestSchema,
    StateCallback,
    DiscoveryCallback
)

class SchemaCallback:
    """
    The SchemaCallback class provides
    interface resources to interact
    with the ST Schema platform from
    an SchemaConnector instance (third-party cloud).

    This class covers:
        - Access Token Request.
        - Refresh Access Tokens.
        - Device State Callback.
        - Device Discovery Callback.
    """
    @classmethod
    def access_token_request(cls, client_id:str, client_secret: str, code: str, refresh_token:str, request_id: str, url: str):
        """
        The access_token_request performs a
        POST Http Requests to the SmartThings
        platform to request or refresh
        Authorization Tokens.
            :::param client_id (required)
            :::param client_secret (required)
            :::param code (required)
            :::param refresh_token (required)
            :::param request_id (required)
            :::param url (required)
        """
        if not url.startswith('https://'):
            raise TypeError('"url" argument not valid, must be Http Secure.')
        else:
            return cls._access_token_request(
                client_id,
                client_secret,
                code,
                refresh_token,
                request_id,
                url
            )

    @staticmethod
    def _access_token_request(*auth_args):
        # Private method that will instance and
        # serialize an AccessTokenRequest object
        # as Request Body to ST Schema OAuth.
        #
        # AccessTokenRequest Instance.
        authentication_data = AccessTokenRequest(
            client_id=auth_args[0],
            client_secret=auth_args[1],
            code=auth_args[2],
            refresh_token=auth_args[3],
            request_id=auth_args[4]
        )
        # Schema Instance and serialization steps.
        schema = AccessTokenRequestSchema()
        authentication_body = schema.dump(authentication_data)
        # POST Http Request to ST Schema
        # OAuth server.
        token_http_request = requests.post(
            url=auth_args[5],  # Url argument
            json=authentication_body
        )
        return token_http_request

    @classmethod
    def state_callback(cls, access_token: str, request_id: str, url: str, devices: list) -> object:
        """
        The state_callback performs a
        POST Http Requests to the SmartThings
        platform to push a new state to the
        devices passed
            :::param access_token (required)
            :::param request_id (required)
            :::param url (required)
            :::param devices (required)
        """
        if not cls._validate_callback_args(url, devices):
            return cls._state_callback(
                access_token,
                request_id,
                url,
                devices
            )

    @staticmethod
    def _state_callback(*callback_args):
        # Private method that will instance and
        # serialize a StateCallback object as
        # Request Body to ST Schema cloud.
        #
        # StateCallback Instance
        state_callback = StateCallback(
            callback_args[0],
            callback_args[1],
            callback_args[3]
        )
        # Schema Instance and serialization steps.
        schema = StateRefreshResponseSchema()
        callback_request_body = schema.dump(state_callback)
        # Post Http Request to ST Schema
        # server.
        return requests.post(
            url=callback_args[2],
            json=callback_request_body
        )

    @classmethod
    def discovery_callback(cls, access_token: str, request_id: str, url: str, devices: list) -> object:
        """
        The discovery_callback performs a
        POST Http Requests to the SmartThings
        platform to push new SchemaDevice
        instances.
            :::param access_token (required)
            :::param request_id (required)
            :::param url (required)
            :::param devices (required)
        """
        if not cls._validate_callback_args(url, devices):
            return cls._discovery_callback(
                access_token,
                request_id,
                url,
                devices
            )

    @staticmethod
    def _discovery_callback(*callback_args):
        # Private method that will instance and
        # serialize a DiscoveryCallback object as
        # Request body to ST Schema Cloud.
        #
        # DiscoveryCallback Instance.
        discovery_callback = DiscoveryCallback(
            access_token=callback_args[0],
            request_id=callback_args[1],
            devices=callback_args[3]
        )
        # Schema Instance and serialization steps
        schema = DiscoveryResponseSchema()
        discovery_callback_body = schema.dump(discovery_callback)
        # POST Http Request to ST Schema
        # server.
        return requests.post(
            url=callback_args[2],
            json=discovery_callback_body
        )

    @staticmethod
    def _validate_callback_args(*callback_args) -> None:
        # Validate Url's protocol
        if not callback_args[0].startswith('https://'):
            raise TypeError('url argument not valid, must be Http Secure.')
        # Validate devices argument type
        if not isinstance(callback_args[1], list):
            raise TypeError('devices argument must be instance of list, not %s' % type(callback_args[1]))
        else:
            # Validate devices items instances
            for device in callback_args[1]:
                if not isinstance(device, SchemaDevice):
                    raise TypeError('devices items must be instance of %s, not %s', (SchemaDevice, type(callback_args[1])))



from marshmallow import Schema, fields, pre_dump, post_dump
from stschema.util.base_modules import HeadersSchema, BaseResponse, AuthenticationSchema
from stschema.schema_device.schemas import DeviceStateSchema, DeviceErrorSchema

from pprint import pprint # Delete line

class StateResponse(BaseResponse):
    """
    Inherits from the BaseResponse class.
    The StateResponse class handles the
    representation of the stateRefreshResponse.

        :::param headers
        :::param device_state
    """

    def __init__(self, devices: list, request_id: str, interaction_type: str = 'stateRefreshResponse'):
        BaseResponse.__init__(self, interaction_type=interaction_type, request_id=request_id)
        self.device_state = devices


class StateRefreshResponseSchema(Schema):
    """
    The StateRefreshResponseSchema handles
    the serialization of the StateResponse
    class supporting the nested
    DeviceStateSchema, HeadersSchema,
    DeviceErrorSchema, and AuthenticationSchema.
    It converts Snake Case attributes
    to Camel Case format following REST
    formatting conventions for JSON
    string objects.
    """

    @pre_dump
    def _verify_dump(self, data, **kwargs):
        # Declare headers as required attribute.
        self.dump_fields.update(
            headers = fields.Nested(HeadersSchema, attribute='headers', required=True))
        # Verify State instance attribute
        # If device_error at device instance,
        # DeviceErrorSchema will be used, else
        # DeviceStateSchema will be used.
        if data.device_state:
            if data.device_state[0].device_error:
                self.dump_fields.update(
                    deviceState=fields.List(fields.Nested(DeviceErrorSchema), attribute='device_state'))
            else:
                self.dump_fields.update(
                    deviceState = fields.List(fields.Nested(DeviceStateSchema), attribute='device_state'))
        # Verify Authentication non-required attribute.
        if data.authentication:
            self.dump_fields.update(
                authentication = fields.Nested(AuthenticationSchema, attribute='authentication'))
        return data

    @post_dump
    def _reset_device_state(self, data, **kwargs):
        # Reset deviceState to support
        # DeviceStateSchema as default
        # schema.
        self.dump_fields.update(
            deviceState = fields.List(fields.Nested(DeviceStateSchema), attribute='device_state'))
        return data



import setuptools
import os


basedir = os.path.abspath(os.path.dirname(__file__))

with open(basedir + '/README.md', 'r') as ld:
    long_description = ld.read()


setuptools.setup(
    name='st-schema-python',
    version='1.0.1',
    author='erickvneri',
    author_email='erickv@bluetrailsoft.com',
    description='SmartThings Schema Connector Python SDK',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/erickvneri/st-schema-python/',
    packages=setuptools.find_packages(),
    install_requires=[
        'pytest',
        'marshmallow',
        'requests',
        'colorama'
    ],
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Home Automation',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ]
)
